import os
import hashlib
import binascii

from fsdicts.lock import Lock

PREFIX_OBJECT = "object-"
PREFIX_REFERENCES = "references-"


class Storage(object):

    def __init__(self, path):
        raise NotImplementedError()

    def put(self, value):
        raise NotImplementedError()

    def readlink(self, link):
        raise NotImplementedError()

    def link(self, identifier, link):
        raise NotImplementedError()

    def unlink(self, link):
        raise NotImplementedError()

    def release(self, identifier):
        raise NotImplementedError()

    def purge(self):
        raise NotImplementedError()


class LinkStorage(Storage):

    def __init__(self, path, hash=hashlib.md5):
        # Make the path absolute
        path = os.path.abspath(path)

        # Intialize the path and hash
        self._path = path
        self._hash = hash

        # Create the path if needed
        if not os.path.exists(self._path):
            os.makedirs(self._path)

    def put(self, value):
        # Create the value hash
        hash = self._hash(value).hexdigest()

        # Create the object path
        path = os.path.join(self._path, hash)

        # Check whether the path exists
        if os.path.isfile(path):
            return path

        # Create temporary path for writing
        temporary = path + "." + binascii.b2a_hex(os.urandom(4)).decode()

        # Write the object
        with open(temporary, "wb") as object:
            object.write(value)

        # Rename the temporary file
        os.rename(temporary, path)

        # Return the hash
        return path

    def readlink(self, link):
        # Read the identifier as a file
        with open(link, "rb") as object:
            value = object.read()

        # Return the value
        return value

    def link(self, identifier, link):
        # Link the identifier and the link
        os.link(identifier, link)

    def unlink(self, link):
        # Check whether the path exists
        if not os.path.isfile(link):
            raise ValueError(link)

        # Read the link and create hash
        with open(link, "rb") as object:
            object_value = object.read()

        # Create object hash
        object_hash = self._hash(object_value).hexdigest()

        # Remove the link
        os.unlink(link)

        # Create object path
        path = os.path.join(self._path, object_hash)

        # Clean the path
        self.release(path)

    def release(self, identifier):
        # Make sure the path exists
        if not os.path.isfile(identifier):
            raise ValueError(identifier)

        # If more then one link exists, skip
        if os.stat(identifier).st_nlink > 1:
            return

        # Remove the file
        os.remove(identifier)

    def purge(self):
        # List all files in the storage and check the link count
        for hash in os.listdir(self._path):
            # Create the file path
            path = os.path.join(self._path, hash)

            # If path does not exist, skip
            if not os.path.isfile(path):
                continue

            # Clean the file
            self.release(path)


class ReferenceStorage(Storage):

    def __init__(self, path, hash=hashlib.md5):
        # Make the path absolute
        path = os.path.abspath(path)

        # Intialize the path and hash
        self._path = path
        self._hash = hash

        # Create the path if needed
        if not os.path.exists(self._path):
            os.makedirs(self._path)

    def put(self, value):
        # Create the value hash
        hash = self._hash(value).hexdigest()

        # Create the object path
        object_path = os.path.join(self._path, PREFIX_OBJECT + hash)

        # Check whether the path exists
        if os.path.isfile(object_path):
            return hash

        # Create temporary path for writing
        temporary = object_path + "." + binascii.b2a_hex(os.urandom(4)).decode()

        # Write the object
        with open(temporary, "wb") as object:
            object.write(value)

        # Rename the temporary file
        os.rename(temporary, object_path)

        # Return the hash
        return hash

    def readlink(self, link):
        # Read the link file
        with open(link, "r") as file:
            identifier = file.read()

        # Create the object path
        object_path = os.path.join(self._path, PREFIX_OBJECT + identifier)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            raise KeyError(identifier)

        # Read the object file
        with open(object_path, "rb") as object:
            value = object.read()

        # Return the value
        return value

    def link(self, identifier, link):
        # Create the object path
        object_path = os.path.join(self._path, PREFIX_OBJECT + identifier)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            raise KeyError(identifier)

        # Create references file path
        references_path = os.path.join(self._path, PREFIX_REFERENCES + identifier)

        # Lock the references
        with Lock(references_path):
            # Append the path to the file
            with open(references_path, "a") as references_file:
                references_file.write(link + "\n")

        # Write the link file
        with open(link, "w") as file:
            file.write(identifier)

    def unlink(self, link):
        # Read the link file
        with open(link, "r") as file:
            identifier = file.read()

        # Create the object path
        object_path = os.path.join(self._path, PREFIX_OBJECT + identifier)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            raise KeyError(identifier)

        # Create references file path
        references_path = os.path.join(self._path, PREFIX_REFERENCES + identifier)

        # Lock the references
        with Lock(references_path):
            # Read the list of references from the file
            with open(references_path, "r") as references_file:
                references = references_file.read().splitlines()

            # Filter out the current reference
            references = list(filter(lambda reference: reference != link, references))

            # Write the list of references back to the file
            with open(references_path, "w") as references_file:
                for reference in references:
                    references_file.write(reference + "\n")

        # Remove the link path
        os.remove(link)

        # Release the object
        self.release(identifier)

    def release(self, identifier):
        # Create the object path
        object_path = os.path.join(self._path, PREFIX_OBJECT + identifier)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            return

        # Create references file path
        references_path = os.path.join(self._path, PREFIX_REFERENCES + identifier)

        # If the references path exists, check references
        if os.path.isfile(references_path):
            # Lock the references
            with Lock(references_path):
                # Read the list of references from the file
                with open(references_path, "r") as references_file:
                    references = references_file.read().splitlines()

                # Filter out references that do not exist
                references = list(filter(lambda reference: os.path.isfile(reference), references))

                # If there are references, return - can't purge
                if references:
                    return

                # Remove the references file
                os.remove(references_path)

        # Remove the object path
        os.remove(object_path)

    def purge(self):
        # List all objects in storage
        for name in os.listdir(self._path):
            # Make sure the name starts with the prefix
            if not name.startswith(PREFIX_OBJECT):
                continue

            # Strip the prefix and release
            identifier = name[len(PREFIX_OBJECT):]

            # Release the object
            self.release(identifier)



import os
import time
import random

# Create lock directory postfix
POSTFIX_LOCK = ".lock"


class Lock(object):

    def __init__(self, path):
        # Create the lock path
        self._path = path + POSTFIX_LOCK

        # Create internal state
        self._locked = False

    def _try_acquire(self):
        try:
            # Try creating the directory
            os.mkdir(self._path)

            # Update lock state
            self._locked = True

            # Locking succeeded
            return True
        except OSError:
            # Locking failed
            return False

    def locked(self):
        return self._locked

    def acquire(self, blocking=True, timeout=None):
        # Mark start time
        start_time = time.time()

        # Try acquiring for the first time
        if self._try_acquire():
            return True

        # If non-blocking, return here
        if not blocking:
            return False

        # Loop until timeout is reached
        while (timeout is None) or (time.time() - start_time) < timeout:
            # Try aquiring the lock
            if self._try_acquire():
                return True

            # Sleep random amount
            time.sleep(random.random() / 1000.0)

        # Timeout reached
        return False

    def release(self):
        # Make sure not already unlocked
        if not self._locked:
            return

        # Try removing the directory
        os.rmdir(self._path)

        # Update the lock status
        self._locked = False

    def __enter__(self):
        # Lock the lock
        self.acquire()

        # Return "self"
        return self

    def __exit__(self, *exc_info):
        # Unlock the lock
        self.release()

    def __str__(self):
        # Create a string representation of the lock
        return "<%s, %s>" % (self.__class__.__name__, "locked" if self._locked else "unlocked")



import os
import hashlib

from fsdicts.bunch import MutableAttributeMapping
from fsdicts.mapping import AdvancedMutableMapping, Mapping

# Create key and value prefixes
PREFIX_KEY = "key-"
PREFIX_VALUE = "value-"

# Create default object so that None can be used as default value
DEFAULT = object()


class Dictionary(AdvancedMutableMapping):

    def __init__(self, path, storage, encoder):
        # Make the path absolute
        path = os.path.abspath(path)

        # Set internal variables
        self._path = path
        self._encode, self._decode = encoder
        self._key_storage, self._value_storage = storage

        # Make sure path exists
        if not os.path.exists(self._path):
            os.makedirs(self._path)

    def _resolve_key(self, key):
        # Hash the key using hashlib
        checksum = hashlib.md5(self._encode(key)).hexdigest()

        # Create both paths
        return os.path.join(self._path, PREFIX_KEY + checksum), os.path.join(self._path, PREFIX_VALUE + checksum)

    def _child_instance(self, path):
        return Dictionary(path, (self._key_storage, self._value_storage), (self._encode, self._decode))

    def __getitem__(self, key):
        # Make sure key exists
        if key not in self:
            raise KeyError(key)

        # Resolve key and value paths
        _, value_path = self._resolve_key(key)

        # Check if object is a simple object
        if os.path.isdir(value_path):
            # Create a keystore from the path
            return self._child_instance(value_path)

        # Read the object value
        encoded_value = self._value_storage.readlink(value_path)

        # Decode the value
        return self._decode(encoded_value)

    def __setitem__(self, key, value):
        # Delete the old value
        if key in self:
            del self[key]

        # Create the key and value paths
        key_path, value_path = self._resolve_key(key)

        # Encode the key
        encoded_key = self._encode(key)

        # Store the key in the object storage
        key_identifier = self._key_storage.put(encoded_key)

        # Link the key to the translation
        self._key_storage.link(key_identifier, key_path)

        # Check if value is a dictionary
        if isinstance(value, Mapping):
            # Create the sub-dictionary
            dictionary = self._child_instance(value_path)

            # Update the dictionary with the values
            dictionary.update(value)

            # Return - nothing more to do
            return

        # Encode the value
        encoded_value = self._encode(value)

        # Store the value in the object storage
        value_identifier = self._value_storage.put(encoded_value)

        # Link the value to the translation
        self._value_storage.link(value_identifier, value_path)

    def __delitem__(self, key):
        # Make sure key exists
        if key not in self:
            raise KeyError(key)

        # Resolve key and value paths
        key_path, value_path = self._resolve_key(key)

        # Release the key object
        self._key_storage.unlink(key_path)

        # If the value is a dictionary, clear the dictionary
        if os.path.isdir(value_path):
            # Create the dictionary
            dictionary = self._child_instance(value_path)

            # Clear the dictionary
            dictionary.clear()

            # Delete the directory - should be empty
            os.rmdir(value_path)

            # Return - nothing more to do
            return

        # Release the value object
        self._value_storage.unlink(value_path)

    def __contains__(self, key):
        # Resolve key and value paths
        key_path, value_path = self._resolve_key(key)

        # Check if paths exist
        return os.path.exists(key_path) and os.path.exists(value_path)

    def __iter__(self):
        # List all of the items in the path
        for name in os.listdir(self._path):
            # Make sure name starts with key prefix
            if not name.startswith(PREFIX_KEY):
                continue

            # Create key path
            key_path = os.path.join(self._path, name)

            # Read the key contents
            encoded_key = self._key_storage.readlink(key_path)

            # Decode the key and yield
            yield self._decode(encoded_key)

    def __len__(self):
        # Count all key files
        return len(list(filter(lambda name: name.startswith(PREFIX_KEY), os.listdir(self._path))))

    def __repr__(self):
        # Format the data like a dictionary
        return "{%s}" % ", ".join("%r: %r" % item for item in self.items())

    def __eq__(self, other):
        # Make sure the other object is a mapping
        if not isinstance(other, Mapping):
            return False

        # Make sure all keys exist
        if set(self.keys()) != set(other.keys()):
            return False

        # Make sure all the values equal
        for key in self:
            if self[key] != other[key]:
                return False

        # Comparison succeeded
        return True

    def pop(self, key, default=DEFAULT):
        try:
            # Fetch the value
            value = self[key]

            # Check if the value is a keystore
            if isinstance(value, Mapping):
                value = value.copy()

            # Delete the item
            del self[key]

            # Return the value
            return value
        except KeyError:
            # Check if a default is defined
            if default != DEFAULT:
                return default

            # Reraise exception
            raise

    def popitem(self):
        # Convert self to list
        keys = list(self)

        # If the list is empty, raise
        if not keys:
            raise KeyError()

        # Pop a key from the list
        key = keys.pop()

        # Return the key and the value
        return key, self.pop(key)

    def copy(self):
        # Create initial bunch
        output = dict()

        # Loop over keys
        for key in self:
            # Fetch value of key
            value = self[key]

            # Check if value is a keystore
            if isinstance(value, Mapping):
                value = value.copy()

            # Update the bunch
            output[key] = value

        # Return the created output
        return output

    def clear(self):
        # Loop over keys
        for key in self:
            # Delete the item
            del self[key]


class AttributeDictionary(Dictionary, MutableAttributeMapping):

    _path = None
    _encode, _decode = None, None
    _key_storage, _value_storage = None, None

    def _child_instance(self, path):
        return AttributeDictionary(path, (self._key_storage, self._value_storage), (self._encode, self._decode))



from fsdicts.lock import *
from fsdicts.bunch import *
from fsdicts.mapping import *
from fsdicts.storage import *
from fsdicts.encoders import *
from fsdicts.database import *
from fsdicts.dictionary import *